@@ -1,6 +1,7 @@
 from __future__ import annotations
 
 import os
+import json
 import ctypes
 import dataclasses
 from typing import Any, Dict, Iterator, List, Optional, Tuple, Union, Protocol
@@ -9,6 +10,8 @@ import llama_cpp.llama as llama
 import llama_cpp.llama_types as llama_types
 import llama_cpp.llama_grammar as llama_grammar
 
+from ._utils import suppress_stdout_stderr
+
 
 class LlamaChatCompletionHandler(Protocol):
     def __call__(
@@ -29,7 +32,7 @@ class LlamaChatCompletionHandler(Protocol):
         response_format: Optional[
             llama_types.ChatCompletionRequestResponseFormat
         ] = None,
-        max_tokens: int = 256,
+        max_tokens: Optional[int] = None,
         presence_penalty: float = 0.0,
         frequency_penalty: float = 0.0,
         repeat_penalty: float = 1.1,
@@ -290,7 +293,7 @@ def register_chat_format(name: str):
             response_format: Optional[
                 llama_types.ChatCompletionRequestResponseFormat
             ] = None,
-            max_tokens: int = 256,
+            max_tokens: Optional[int] = None,
             presence_penalty: float = 0.0,
             frequency_penalty: float = 0.0,
             repeat_penalty: float = 1.1,
@@ -316,10 +319,11 @@ def register_chat_format(name: str):
                 stop = [] if stop is None else [stop] if isinstance(stop, str) else stop
                 rstop = result.stop if isinstance(result.stop, list) else [result.stop]
                 stop = stop + rstop
-            
+
             if response_format is not None and response_format["type"] == "json_object":
-                print("hello world")
-                grammar = llama_grammar.LlamaGrammar.from_string(llama_grammar.JSON_GBNF)
+                grammar = llama_grammar.LlamaGrammar.from_string(
+                    llama_grammar.JSON_GBNF
+                )
 
             completion_or_chunks = llama.create_completion(
                 prompt=prompt,
@@ -561,7 +565,7 @@ def format_chatml(
     _messages = _map_roles(messages, _roles)
     _messages.append((_roles["assistant"], None))
     _prompt = _format_chatml(system_message, _messages, _sep)
-    return ChatFormatterResponse(prompt=_prompt)
+    return ChatFormatterResponse(prompt=_prompt, stop=_sep)
 
 
 @register_chat_completion_handler("functionary")
@@ -570,12 +574,15 @@ def functionary_chat_handler(
     messages: List[llama_types.ChatCompletionRequestMessage],
     functions: Optional[List[llama_types.ChatCompletionFunction]] = None,
     function_call: Optional[llama_types.ChatCompletionRequestFunctionCall] = None,
+    tools: Optional[List[llama_types.ChatCompletionTool]] = None,
+    tool_choice: Optional[llama_types.ChatCompletionToolChoiceOption] = None,
     temperature: float = 0.2,
     top_p: float = 0.95,
     top_k: int = 40,
     stream: bool = False,
     stop: Optional[Union[str, List[str]]] = [],
-    max_tokens: int = 256,
+    response_format: Optional[llama_types.ChatCompletionRequestResponseFormat] = None,
+    max_tokens: Optional[int] = None,
     presence_penalty: float = 0.0,
     frequency_penalty: float = 0.0,
     repeat_penalty: float = 1.1,
@@ -590,57 +597,80 @@ def functionary_chat_handler(
 ) -> Union[llama_types.ChatCompletion, Iterator[llama_types.ChatCompletionChunk]]:
     SYSTEM_MESSAGE = """A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions. The assistant calls functions with appropriate input when necessary"""
 
-    def generate_schema_from_functions(
-        functions: List[llama_types.ChatCompletionFunctions],
-        namespace: str = "functions",
-    ):
-        """
-        Convert functions schema to a schema that language models can understand.
-        """
+    def generate_type_definition(param: Dict[str, llama_types.JsonType], indent_level: int, shared_defs) -> str:
+        indent = '  ' * indent_level
+        if '$ref' in param:
+            # Reference to a shared definition
+            ref_name = param['$ref'].split('/')[-1]  # Extract the type name from the reference
+            return ref_name
+        elif param.get('type') == 'array':
+            items = param.get('items', {})
+            item_type = generate_type_definition(items, indent_level + 1, shared_defs)
+            return f"Array<{item_type}>"
+        elif param.get('type') == 'object':
+            properties = param.get('properties', {})
+            nested_schema = "{\n"
+            for nested_param_name, nested_param in properties.items():
+                nested_param_type = generate_type_definition(nested_param, indent_level + 1, shared_defs)
+                nested_schema += f"{indent}  {nested_param_name}: {nested_param_type},\n"
+            nested_schema += indent + "}"
+            return nested_schema
+        elif 'enum' in param:
+            # Enum type
+            return " | ".join([f'"{enum_value}"' for enum_value in param['enum']])
+        else:
+            # Simple type
+            return param.get('type', 'any')
 
-        schema = (
-            "// Supported function definitions that should be called when necessary.\n"
-        )
+    def generate_shared_definitions(shared_defs, indent_level: int) -> str:
+        indent = '  ' * indent_level
+        shared_definitions = ""
+        for def_name, def_properties in shared_defs.items():
+            shared_definitions += f"{indent}type {def_name} = "
+            if def_properties.get('type') == 'object':
+                shared_definitions += generate_type_definition(def_properties, indent_level, shared_defs)
+            elif 'enum' in def_properties:
+                # Enum type
+                shared_definitions += " | ".join([f'"{enum_value}"' for enum_value in def_properties['enum']])
+            shared_definitions += ";\n"
+        return shared_definitions
+
+    def generate_schema_from_functions(functions, namespace="functions") -> str:
+        schema = "// Supported function definitions that should be called when necessary.\n"
         schema += f"namespace {namespace} {{\n\n"
 
+        # Generate shared definitions
+        shared_definitions = {}
+        for function in functions:
+            parameters = function.get("parameters", {})
+            shared_definitions.update(parameters.get("$defs", {}))
+
+        schema += generate_shared_definitions(shared_definitions, 1)
+
         for function in functions:
-            # Convert a Function object to dict, if necessary
             function_name = function["name"]
             description = function.get("description", "")
-            schema += f"// {description}\n"
-            schema += f"type {function_name}"
-
-            parameters = function.get("parameters", None)
-            schema += " = (_: {\n"
+            parameters = function.get("parameters", {})
             required_params = parameters.get("required", [])
+            
+            schema += f"  // {description}\n"
+            schema += f"  type {function_name} = (_: {{\n"
+            
             for param_name, param in parameters.get("properties", {}).items():
-                # Param Description
-                description = param.get("description")
-                if description is not None:
-                    schema += f"// {description}\n"
-
-                # Param Name
-                schema += f"{param_name}"
-                if param_name not in required_params:
-                    schema += "?"
-
-                # Param Type
-                param_type = param.get("type", "any")
-                if param_type == "integer":
-                    param_type = "number"
-                if "enum" in param:
-                    param_type = " | ".join([f'"{v}"' for v in param["enum"]])
-                schema += f": {param_type},\n"
-
-            schema += "}) => any;\n\n"
-
-        schema += f"}} // namespace {namespace}"
+                param_description = param.get("description", "")
+                param_type = generate_type_definition(param, 2, shared_definitions)
+                optional_indicator = "" if param_name in required_params else "?"
+                schema += f"    // {param_description}\n"
+                schema += f"    {param_name}{optional_indicator}: {param_type},\n"
+            schema += "  }) => any;\n\n"
 
+        schema += "}} // namespace {}\n".format(namespace)
         return schema
 
     def prepare_messages_for_inference(
         messages: List[llama_types.ChatCompletionRequestMessage],
         functions: Optional[List[llama_types.ChatCompletionFunctions]] = None,
+        tools: Optional[List[llama_types.ChatCompletionTool]] = None,
     ):
         all_messages: List[llama_types.ChatCompletionRequestMessage] = []
         if functions is not None:
@@ -649,6 +679,15 @@ def functionary_chat_handler(
                     role="system", content=generate_schema_from_functions(functions)
                 )
             )
+        
+        if tools is not None:
+            all_messages.append(
+                llama_types.ChatCompletionRequestSystemMessage(
+                    role="system", content=generate_schema_from_functions(
+                        [tool["function"] for tool in tools if tool["type"] == "function"]
+                    )
+                )
+            )
 
         all_messages.append(
             llama_types.ChatCompletionRequestSystemMessage(
@@ -681,16 +720,24 @@ def functionary_chat_handler(
                 return f"function name={msg['name']}:\n{msg['content']}\n"
             elif msg["role"] == "function" and "function_call" in msg:
                 return f"function name={msg['function_call']['name']}:\n{msg['function_call']['arguments']}\n"
+            elif msg["role"] == "tool":
+                if msg["content"] is not None:
+                    return f"function name={msg['tool_call_id']}:\n{msg['content']}\n"
+                else:
+                    return f"function name={msg['tool_call_id']}\n"
             elif msg["role"] == "user":
                 if msg["content"] is None:
-                    return "user:\n</s>"
+                    return "user:\n</s></s>\n"
                 else:
-                    return f"user:\n</s>{msg['content']}\n"
+                    return f"user:\n</s>{msg['content']}</s>\n"
             elif msg["role"] == "assistant":
                 if msg["content"] is not None and "function_call" in msg:
-                    return f"assistant:\n{msg['content']}\nassistant to={msg['function_call']['name']}:\n{msg['function_call']['arguments']}</s>"
+                    return f"assistant:\n{msg['content']}\nassistant to={msg['function_call']['name']}:\n{msg['function_call']['arguments']}</s>\n"
                 elif "function_call" in msg:
-                    return f"assistant to={msg['function_call']['name']}:\n{msg['function_call']['arguments']}</s>"
+                    return f"assistant to={msg['function_call']['name']}:\n{msg['function_call']['arguments']}</s>\n"
+                elif "tool_calls" in msg and len(msg["tool_calls"]) > 0:
+                    for tool_call in msg["tool_calls"]: # NOTE: probably doesn't work with the functionary model
+                        return f"assistant to={tool_call['id']}:\n{tool_call['function']['arguments']}</s>\n"
                 elif msg["content"] is None:
                     return "assistant"
                 else:
@@ -699,8 +746,14 @@ def functionary_chat_handler(
                 raise ValueError(f"Unsupported role: {msg['role']}")
 
         return "".join([message_to_str(msg) for msg in all_messages])
+    
+    if tools is not None:
+        functions = [tool["function"] for tool in tools if tool["type"] == "function"]
+    
+    if tool_choice is not None:
+        function_call = tool_choice if isinstance(tool_choice, str) else tool_choice["function"]
 
-    prompt = prepare_messages_for_inference(messages, functions)
+    prompt = prepare_messages_for_inference(messages, functions, tools)
 
     if function_call is None and (functions is None or len(functions) == 0):
         completion_or_completion_chunks = llama.create_completion(
@@ -733,24 +786,69 @@ def functionary_chat_handler(
         )  # type: ignore
         completion_text = completion["choices"][0]["text"]
         # strip " to=functions." and ending ":"
-        function_call = completion_text[14:-1]
+        function_call = completion_text.split(".")[-1][:-1]
         new_prompt = prompt + completion_text + stop
     elif isinstance(function_call, str) and function_call != "none":
-        new_prompt = prompt + f"assistant:\n"
+        new_prompt = prompt + f":\n"
     elif isinstance(function_call, dict):
-        new_prompt = prompt + f"assistant to={function_call['name']}:\n"
+        new_prompt = prompt + f" to=functions.{function_call['name']}:\n"
         function_call = function_call["name"]
     else:
-        new_prompt = prompt + f"assistant:\n"
+        new_prompt = prompt + f":\n"
+
+    function_body = None
+    for function in functions or []:
+        if function["name"] == function_call:
+            function_body = function["parameters"]
+            break
+    for tool in tools or []:
+        if tool["type"] == "function" and tool["function"]["name"] == function_call:
+            function_body = tool["function"]["parameters"]
+            break
+    
+    if function_body is not None:
+        try:
+            with suppress_stdout_stderr(disable=llama.verbose):
+                grammar_text = llama_grammar.json_schema_to_gbnf(json.dumps(function_body))
+                grammar = llama_grammar.LlamaGrammar.from_string(llama_grammar.json_schema_to_gbnf(json.dumps(function_body)))
+                print(grammar_text)
+        except Exception as e:
+            if llama.verbose:
+                print("Failed to parse function body as JSON schema, falling back to default grammar")
+                print(e)
+            with suppress_stdout_stderr(disable=llama.verbose):
+                grammar = llama_grammar.LlamaGrammar.from_string(llama_grammar.JSON_GBNF)
+    else:
+        with suppress_stdout_stderr(disable=llama.verbose):
+            grammar = llama_grammar.LlamaGrammar.from_string(llama_grammar.JSON_GBNF)
 
     completion: llama_types.Completion = llama.create_completion(
-        prompt=new_prompt, stop=["user:", "</s>"], stream=False
+        prompt=new_prompt,
+        stop=["user:", "</s>"],
+        stream=False,
+        grammar=grammar,
+        max_tokens=max_tokens,
+        temperature=temperature,
+        top_p=top_p,
+        top_k=top_k,
+        presence_penalty=presence_penalty,
+        frequency_penalty=frequency_penalty,
+        repeat_penalty=repeat_penalty,
+        tfs_z=tfs_z,
+        mirostat_mode=mirostat_mode,
+        mirostat_tau=mirostat_tau,
+        mirostat_eta=mirostat_eta,
+        model=model,
+        logits_processor=logits_processor,
     )  # type: ignore
 
     assert "usage" in completion
     assert isinstance(function_call, str)
     assert stream is False  # TODO: support stream mode
 
+    print(new_prompt)
+    print(completion["choices"][0]["text"])
+
     return llama_types.CreateChatCompletionResponse(
         id="chat" + completion["id"],
         object="chat.completion",
@@ -760,14 +858,24 @@ def functionary_chat_handler(
             {
                 "index": 0,
                 "message": {
-                    "role": "function",
+                    "role": "assistant",
                     "content": None,
                     "function_call": {
                         "name": function_call,
                         "arguments": completion["choices"][0]["text"],
                     },
+                    "tool_calls": [
+                        {
+                            "id": function_call,
+                            "type": "function",
+                            "function": {
+                                "name": function_call,
+                                "arguments": completion["choices"][0]["text"],
+                            }
+                        }
+                    ]
                 },
-                "finish_reason": "function_call",
+                "finish_reason": "tool_calls",
             }
         ],
         usage=completion["usage"],
@@ -775,20 +883,26 @@ def functionary_chat_handler(
 
 
 class Llava15ChatHandler:
-    def __init__(self, clip_model_path: str):
+    _clip_free = None
+
+    def __init__(self, clip_model_path: str, verbose: bool = False):
         import llama_cpp.llava_cpp as llava_cpp
 
         self._llava_cpp = llava_cpp
         self.clip_model_path = clip_model_path
+        self.verbose = verbose
+        self._clip_free = self._llava_cpp._libllava.clip_free  # type: ignore
 
-        self.clip_ctx = self._llava_cpp.clip_model_load(
-            self.clip_model_path.encode(), 0
-        )
+        with suppress_stdout_stderr(disable=self.verbose):
+            self.clip_ctx = self._llava_cpp.clip_model_load(
+                self.clip_model_path.encode(), 0
+            )
 
     def __del__(self):
-        if self.clip_ctx is not None:
-            self._llava_cpp.clip_free(self.clip_ctx)
-            self.clip_ctx = None
+        with suppress_stdout_stderr(disable=self.verbose):
+            if self.clip_ctx is not None and self._clip_free is not None:
+                self._clip_free(self.clip_ctx)
+                self.clip_ctx = None
 
     def load_image(self, image_url: str) -> bytes:
         if image_url.startswith("data:"):
@@ -817,7 +931,10 @@ class Llava15ChatHandler:
         top_k: int = 40,
         stream: bool = False,
         stop: Optional[Union[str, List[str]]] = [],
-        max_tokens: int = 256,
+        response_format: Optional[
+            llama_types.ChatCompletionRequestResponseFormat
+        ] = None,
+        max_tokens: Optional[int] = None,
         presence_penalty: float = 0.0,
         frequency_penalty: float = 0.0,
         repeat_penalty: float = 1.1,
@@ -843,7 +960,6 @@ class Llava15ChatHandler:
             if system_prompt != ""
             else "A chat between a curious human and an artificial intelligence assistant.  The assistant gives helpful, detailed, and polite answers to the human's questions."
         )
-        system_prompt = "A chat between a curious human and an artificial intelligence assistant.  The assistant gives helpful, detailed, and polite answers to the human's questions."
         user_role = "\nUSER:"
         assistant_role = "\nASSISTANT:"
         llama.reset()
@@ -881,36 +997,47 @@ class Llava15ChatHandler:
                             c_ubyte_ptr = (
                                 ctypes.c_ubyte * len(data_array)
                             ).from_buffer(data_array)
-                            embed = self._llava_cpp.llava_image_embed_make_with_bytes(
-                                ctx_clip=self.clip_ctx,
-                                n_threads=llama.context_params.n_threads,
-                                image_bytes=c_ubyte_ptr,
-                                image_bytes_length=len(image_bytes),
-                            )
-                            # image_bytes_p = (ctypes.c_uint8 * len(image_bytes)).from_buffer_copy(image_bytes)
-                            # embed = self._llava_cpp.llava_image_embed_make_with_bytes(ctx_clip=self.clip_ctx, n_threads=1, image_bytes=image_bytes_p, image_bytes_length=len(image_bytes))
+                            with suppress_stdout_stderr(disable=self.verbose):
+                                embed = (
+                                    self._llava_cpp.llava_image_embed_make_with_bytes(
+                                        ctx_clip=self.clip_ctx,
+                                        n_threads=llama.context_params.n_threads,
+                                        image_bytes=c_ubyte_ptr,
+                                        image_bytes_length=len(image_bytes),
+                                    )
+                                )
                             try:
                                 n_past = ctypes.c_int(llama.n_tokens)
                                 n_past_p = ctypes.pointer(n_past)
-                                self._llava_cpp.llava_eval_image_embed(
-                                    ctx_llama=llama.ctx,
-                                    embed=embed,
-                                    n_batch=llama.n_batch,
-                                    n_past=n_past_p,
-                                )
+                                with suppress_stdout_stderr(disable=self.verbose):
+                                    self._llava_cpp.llava_eval_image_embed(
+                                        ctx_llama=llama.ctx,
+                                        embed=embed,
+                                        n_batch=llama.n_batch,
+                                        n_past=n_past_p,
+                                    )
                                 assert llama.n_ctx() >= n_past.value
                                 llama.n_tokens = n_past.value
                             finally:
-                                self._llava_cpp.llava_image_embed_free(embed)
+                                with suppress_stdout_stderr(disable=self.verbose):
+                                    self._llava_cpp.llava_image_embed_free(embed)
             if message["role"] == "assistant" and message["content"] is not None:
                 llama.eval(
                     llama.tokenize(
                         f"ASSISTANT: {message['content']}".encode("utf8"), add_bos=False
                     )
                 )
+                assert llama.n_ctx() >= llama.n_tokens
         llama.eval(llama.tokenize(f"{assistant_role}".encode("utf8"), add_bos=False))
+        assert llama.n_ctx() >= llama.n_tokens
 
-        prompt = llama._input_ids.tolist()
+        prompt = llama.input_ids[: llama.n_tokens].tolist()
+
+        if response_format is not None and response_format["type"] == "json_object":
+            with suppress_stdout_stderr(disable=self.verbose):
+                grammar = llama_grammar.LlamaGrammar.from_string(
+                    llama_grammar.JSON_GBNF
+                )
 
         return _convert_completion_to_chat(
             llama.create_completion(
